--****************************************************************************
--**
--**  File     :  /lua/unit.lua
--**  Author(s):  John Comes, David Tomandl, Gordon Duclos
--**
--**  Summary  : The Unit lua module
--**
--**  Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
--****************************************************************************

local oldUnit=Unit
Unit = Class(oldUnit) {

    OnCreate = function(self)
        oldUnit.OnCreate(self)
        local bp = self:GetBlueprint()
        if bp.Economy.XPperLevel then -- gh, just to make certain this is not called
            --for units which do not need it
            self.XPnextLevel = bp.Economy.XPperLevel
            self.xp = 0
--            self.Txp = 0 --gh, disabled, ui code works fine
            self.VeteranLevel = 1
            self.LevelProgress = 1
            -- gh, removed BuildRate Sync Entry in order to speed up gameplay, there's an engine method we will use now.
--            self.Sync.XPnextLevel = self.XPnextLevel --Two new sync entries
--            self.Sync.Txp = self.Txp --was slowing things down
            self.Sync.LevelProgress = self.LevelProgress
            self.Sync.RegenRate = bp.Defense.RegenRate
        end
    end,

    OnStopBeingBuilt = function(self, builder, layer)
        if self:GetBlueprint().Economy.xpTimeStep and EntityCategoryContains(categories.ENERGYPRODUCTION, self) then
            self:ForkThread(self.XPOverTime)
        end--gh, only triggers energy XPOverTime, fixes threading issue
        oldUnit.OnStopBeingBuilt(self, builder, layer)
    end,


    --Economy
    --gh, this is for everything which requires energy to be run: intel, shields etc
    SetMaintenanceConsumptionActive = function(self)
        if self.XPOverTimeThread then KillThread(self.XPOverTimeThread) end
        self.XPOverTimeThread = ForkThread(self.XPOverTime, self)
        oldUnit.SetMaintenanceConsumptionActive(self)
    end,

    XPOverTime = function(self)
        if not self:GetBlueprint().Economy.xpTimeStep then return end -- sanity check
        local waittime = self:GetBlueprint().Economy.xpTimeStep * 0.5
        WaitSeconds(waittime) --gh, wait more to improve performance
        while not self:IsDead() and (self.MaintenanceConsumption or EntityCategoryContains(categories.ENERGYPRODUCTION, self))do
            self:AddXP(self.XPnextLevel* 0.5)
            WaitSeconds(waittime)
        end
    end,

    -- gh, reworked builder xp to include upgrades and allow for pausing
    startBuildXPThread = function(self)
        local levelPerSecond = self:GetBlueprint().Economy.BuildXPLevelpSecond
        if not levelPerSecond then return end
        WaitSeconds(1)--gh, when you build THAT fast you do not need any xp, reduces autopausing woes
        while not self:IsDead() and self.ActiveConsumption do
            --gh, check added, now this loop kills itself if there is no consumption
            self:AddXP(self.XPnextLevel*levelPerSecond*0.01)
            WaitSeconds(59)
        end
    end,

    SetActiveConsumptionActive = function(self)
        if self.BuildXPThread then KillThread(self.BuildXPThread) end
        self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
        oldUnit.SetActiveConsumptionActive(self)
    end,
--this applied to builders

--Nukes/Silos
    startSiloXPThread = function(self)
        local levelPerSecond = self:GetBlueprint().Economy.BuildXPLevelpSecond
        if not levelPerSecond then return end
        WaitSeconds(1)
        while not self:IsDead() and self:IsUnitState("SiloBuildingAmmo") do
            --gh, added check if silo is building
            self:AddXP(self.XPnextLevel*levelPerSecond*0.01)
            WaitSeconds(59)
        end
    end,--gh, seperate thread, we do not want this to interfere on ACUs and SCUs

    OnSiloBuildStart = function(self, weapon)
        if self.SiloXPThread then KillThread(self.SiloXPThread) end
        self.SiloXPThread = ForkThread(self.startSiloXPThread, self)
        oldUnit.OnSiloBuildStart(self, weapon)
    end,



    ------------------------------------------------------------------------------------- VETERANCY
    ---------------------------------------------------------------------------------

    -- use this to go through the AddKills function rather than directly setting veterancy
    SetVeterancy = function(self, veteranLevel)
        veteranLevel = veteranLevel or 0
        if veteranLevel <= 5 then
            return oldUnit.SetVeterancy(self, veteranLevel)
        else
            local bp = self:GetBlueprint()-- change suggested by Gilbot
            if bp.Veteran['Level'..veteranLevel] then
                self:AddKills(bp.Veteran['Level'..veteranLevel])
            else
                WARN('SetVeterancy called on ' .. self:GetUnitId()
                .. ' with veteran level ' .. veteranLevel
                .. ' which was not defined in its BP file. '
                .. ' Veterancy level has not been set.'
                )
            end
        end
    end,

    --Check to see if we should veteran up.
    CheckVeteranLevel = function(self)
        if not self.XPnextLevel then return end
        local bp = self:GetBlueprint()
--        LOG('xp:' .. self.xp.. ' level at:' .. self.XPperLevel)
        while self.xp >= self.XPnextLevel do
            self.xp = self.xp - self.XPnextLevel
            self.XPnextLevel = bp.Economy.XPperLevel * (1+ 0.1*self.VeteranLevel)
            self:SetVeteranLevel(self.VeteranLevel + 1)--gh, fix calling this so often
            --takes too much time
        end
--        if self.xp >= self.XPnextLevel then
--            local count = 0
--            while self.xp >= self.XPnextLevel do
--                self.xp = self.xp - self.XPnextLevel
--                local curLevel = self.VeteranLevel+count
--                self.XPnextLevel = bp.Economy.XPperLevel * (1+  0.1*curLevel)
--                count = count + 1
--            end
--            self:SetVeteranLevel(count+self.VeteranLevel)
--        end --gh, if i ever fix the buff. lua ...
        self.LevelProgress = self.xp / self.XPnextLevel + self.VeteranLevel
        self.Sync.LevelProgress = self.LevelProgress
--        self.Sync.XPnextLevel = self.Txp - self.xp + self.XPnextLevel
        --gh, syncing XP needed for Levelup, calculation within the UI may have minor errors
    end,

--gh, here's the gpg function fixed, lightweight, which is good
--but not very noticeable at low levels
--consider some better solution
    PlayVeteranFx = function(self,time)
        time = time or 1
        CreateAttachedEmitter(self, 0, self:GetArmy(), '/effects/emitters/destruction_explosion_concussion_ring_01_emit.bp'):ScaleEmitter(time)
    end,

    --Set the veteran level to the level specified
    --gh, fixme: Only call this function once on level gain > 1
    SetVeteranLevel = function(self, level)

        --LOG('*DEBUG: '.. self:GetBlueprint().Description .. ' VETERAN UP! LEVEL ', repr(level))
        --local cnt = level - self.VeteranLevel
        self.VeteranLevel = level

        -- Apply default veterancy buffs
        local buffTypes = {
            'Health',
            'Regen',
            'Vision',
            'Damage',
            'DamageArea',
            'Range',
            'RateOfFire',
            'Speed',
            'BuildRate',
            'Radar',
            'Sonar',
            'OmniRadius',
            'ResourceProduction',
            'Shield'
        }-- gh, can be removed

        --Ghaleon
        --we check if a unit needs a buff, default applied all buffs on any unit levelup
        --Took massive amounts of time to do and possibly inflated ram usage much
        Buff.ApplyBuff( self, 'VeterancyHealth')
        Buff.ApplyBuff( self, 'VeterancyRegen')
        Buff.ApplyBuff( self, 'VeterancyVision')
        local bp = self:GetBlueprint()
        local gotweapon = false
        if bp.Weapon then
            for k,v in bp.Weapon do
                if v.Label != 'DeathWeapon' then
--                    LOG(repr(v.Label))
                    Buff.ApplyBuff( self, 'VeterancyDamage')
                    Buff.ApplyBuff( self, 'VeterancyDamageArea')
                    Buff.ApplyBuff( self, 'VeterancyRange')
                    Buff.ApplyBuff( self, 'VeterancyRateOfFire')
                    gotweapon = true
                    break
                end
            end
        end
        if EntityCategoryContains(categories.MOBILE, self) and not EntityCategoryContains(categories.AIR, self) then
            Buff.ApplyBuff( self, 'VeterancySpeed')
        end
        if self:GetBuildRate() and self:GetBuildRate() > 2 then
            Buff.ApplyBuff( self, 'VeterancyBuildRate')
        end
        if bp.Intel.RadarRadius and bp.Intel.RadarRadius > 0 then
            Buff.ApplyBuff( self, 'VeterancyRadar')
        end
        if bp.Intel.SonarRadius and bp.Intel.SonarRadius > 0 then
            Buff.ApplyBuff( self, 'VeterancySonar')
        end
        if bp.Intel.OmniRadius and bp.Intel.OmniRadius > 0 then
            Buff.ApplyBuff( self, 'VeterancyOmniRadius')
        end
        if (bp.Economy.ProductionPerSecondEnergy and bp.Economy.ProductionPerSecondEnergy > 0) or (bp.Economy.ProductionPerSecondMass and bp.Economy.ProductionPerSecondMass > 0) then
            Buff.ApplyBuff( self, 'VeterancyResourceProduction')
        end
        local gotShield = false
        if EntityCategoryContains(categories.COMMAND, self) or EntityCategoryContains(categories.SUBCOMMANDER, self) then

            Buff.ApplyBuff( self, 'VeterancyCommandProduction')

            Buff.ApplyBuff( self, 'VeterancySpeed2')

            if not EntityCategoryContains(categories.CYBRAN, self) then
                gotShield = true
            end
        end--gh, here we got our ACU and SCU buffs
        if self:GetShield() or gotShield then
            Buff.ApplyBuff( self, 'VeterancyShield')
        end
        -- Get any overriding buffs if they exist
        local bpb = self:GetBlueprint().Buffs
        --Check for unit buffs

        if bpb then
            for bLevel,bData in bpb do
                if (bLevel == 'Any' or bLevel == 'Level'..level) then
                    --gh, kept 'Any' as the Soothsayer *might* need it, do not care enough atm
                    --need to check if i can fully eliminate this one
                    for bType,bValues in bData do
                        local buffName = self:CreateUnitBuff(bLevel,bType,bValues)
                        if buffName then
                            Buff.ApplyBuff( self, buffName )
                        end
                    end
                end
            end
        end
        if gotweapon then -- gh, the Fx for units
            if level == 6 then self:PlayVeteranFx(0.5) end
            if math.mod(level,10) == 1 then self:PlayVeteranFx(level*0.1) end
        end
        self:GetAIBrain():OnBrainUnitVeterancyLevel(self, level)
        self:DoUnitCallbacks('OnVeteran')
    end,

    --function to generate the new veterancy buffs
    --gh, do we really need this ?
    CreateUnitBuff = function(self, levelName, buffType, buffValues)

        -- Generate a buff based on the unitId
        local buffName = self:GetUnitId() .. levelName .. buffType
        local buffMinLevel = nil
        local buffMaxLevel = nil
        if buffValues.MinLevel then buffMinLevel = buffValues.MinLevel end
        if buffValues.MaxLevel then buffMaxLevel = buffValues.MaxLevel end


        -- Create the buff if needed
        if not Buffs[buffName] then
            --LOG(buffName .. ': '..buffMinLevel.. ' - '..buffMaxLevel)
            BuffBlueprint {
                MinLevel = buffMinLevel,
                MaxLevel = buffMaxLevel,
                Name = buffName,
                DisplayName = buffName,
                BuffType = buffType,
                Stacks = buffValues.Stacks,
                --self.BuffTypes[buffType].BuffStacks,
                Duration = buffValues.Duration,
                Affects = buffValues.Affects,
            }
        end

        -- Return the buffname so the buff can be applied to the unit
        return buffName
    end,
    --checkthis

    --Allowing buff bonus and adj bonus at same time!!
    UpdateProductionValues = function(self)
        local bpEcon = self:GetBlueprint().Economy
        if not bpEcon then return end
        self:SetProductionPerSecondEnergy( (self.EnergyProdMod or bpEcon.ProductionPerSecondEnergy or 0)* (self.EnergyProdAdjMod or 1) )
        self:SetProductionPerSecondMass( (self.MassProdMod or bpEcon.ProductionPerSecondMass or 0) * (self.MassProdAdjMod or 1) )
    end,

    AddLevels = function(self, levels)
        local bp = self:GetBlueprint()
        local curlevel = self.VeteranLevel
        local percent = self.LevelProgress - curlevel
        local xpAdd = 0
        if levels >= (1-percent) then
            xpAdd = self.XPnextLevel * (1-percent)
            levels=levels-(1-percent)
        else
            xpAdd =self.XPnextLevel * levels
            levels=0
        end
        while levels > 1 do
            levels=levels-1
            curlevel = curlevel +1
            xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.1*curlevel)
        end
        xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.1*(curlevel+1)) * levels
        self:AddXP(xpAdd) --same
    end,

    AddXP = function(self,amount)
        if not self.XPnextLevel then return end
--        self.Txp = self.Txp + (math.ceil(amount))--slowed things down
--        self.Sync.Txp = self.Txp-- Total EXP collected
        self.xp = self.xp + (amount)--removed ceil, hope this speeds things up
--        LOG('___' .. amount .. repr(self:GetBlueprint().Description))
        self:CheckVeteranLevel()
    end,

    --gh, disabled, helps simspeed to not check things on every hit
--    DoTakeDamage = function(self, instigator, amount, vector, damageType)
--        if instigator and IsUnit(instigator) and not instigator:IsDead() then
--            local bp = self:GetBlueprint()
--            if bp.Economy.xpPerHp then
--                local preAdjHealth = self:GetHealth()
--                --check if damage kills self and if self has positive health
--                if amount >= preAdjHealth and preAdjHealth > 0 then
--                    local sLevel = self.VeteranLevel
----                    LOG('___sL: ' .. sLevel)
--                    local xpR = sLevel
--                    instigator:AddXP(bp.Economy.xpValue*xpR*0.25)
--                    --xp gains only on killing a unit
----                    LOG('___xpR: ' .. xpR  )
----                    LOG('___xpV: ' .. bp.Economy.xpValue  )
----                    LOG('___xpT: ' .. (bp.Economy.xpValue*xpR*AoE))
--                end
--            end
--        end
--        oldUnit.DoTakeDamage(self, instigator, amount, vector, damageType)
--    end,
    --gh,
    OnKilled = function(self, instigator, type, overkillRatio)
        if instigator and IsUnit(instigator) and not instigator:IsDead() then
            local bp = self:GetBlueprint()
            if bp.Economy.xpPerHp then
                local sLevel = self.VeteranLevel
                local xpR = sLevel --just to quickly adjust this calculation
                instigator:AddXP(bp.Economy.xpValue*xpR*0.25)
            end
        end
        oldUnit.OnKilled(self, instigator, type, overkillRatio)
    end,


--disables friendly fire, projectiles no longer damage or hit allies
-- caused projectiles to NEVER hit, disabled.
--    OnCollisionCheck = function(self, other, firingWeapon)
--    if IsAlly(self:GetArmy(), other:GetArmy()) then return false end
--    --credit:Moolio (thanks!)
--    old.OnCollisionCheck(self, other, firingWeapon)
--    end,


--SetBuildRange = function(self,amount)
----dummy, if arbitrary blueprint stats can be changed
--end,

    InitiateTeleportThread = function(self, teleporter, location, orientation)
        local tbp = teleporter:GetBlueprint()
        local ubp = self:GetBlueprint()
        self.UnitBeingTeleported = self
        self:SetImmobile(true)
        self:PlayUnitSound('TeleportStart')
        self:PlayUnitAmbientSound('TeleportLoop')
        local bp = self:GetBlueprint().Economy
        local energyCost, time
        if bp then
            local mass = bp.BuildCostMass * (bp.TeleportMassMod or 0.01)
            local energy = bp.BuildCostEnergy * (bp.TeleportEnergyMod or 0.01)
            energyCost = mass + energy
            time = energyCost * (bp.TeleportTimeMod or 0.01)
            if (time - self.VeteranLevel) > 0 then
                time = time - self.VeteranLevel
            else time = 0.1 end
                -- gh, veterans teleport faster
                -- too bad this can't be done with less code
            end

            --LOG('*UNIT DEBUG: TELEPORTINgh, energyCost = ', repr(energyCost), ' time = ', repr(time))
            self.TeleportDrain = CreateEconomyEvent(self, energyCost or 100, 0, time or 5, self.UpdateTeleportProgress)

            -- create teleport charge effect
            self:PlayTeleportChargeEffects()

            WaitFor( self.TeleportDrain  ) -- Perform fancy Teleportation FX here

            if self.TeleportDrain then
                RemoveEconomyEvent(self, self.TeleportDrain )
                self.TeleportDrain = nil
            end

            self:PlayTeleportOutEffects()
            self:CleanupTeleportChargeEffects()

            WaitSeconds( 0.1 )

            self:SetWorkProgress(0.0)
            Warp(self, location, orientation)
            self:PlayTeleportInEffects()

            WaitSeconds( 0.1 ) -- Perform cooldown Teleportation FX here
            --Landing Sound
            --LOG('DROP')
            self:StopUnitAmbientSound('TeleportLoop')
            self:PlayUnitSound('TeleportEnd')
            self:SetImmobile(false)
            self.UnitBeingTeleported = nil
            self.TeleportThread = nil
        end,


        GetShield = function(self)
            return self.MyShield or nil
        end,
    }
